/* Intel Silvermont gmp-mparam.h -- Compiler/machine parameter header file.

Copyright 2019 Free Software Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of either:

  * the GNU Lesser General Public License as published by the Free
    Software Foundation; either version 3 of the License, or (at your
    option) any later version.

or

  * the GNU General Public License as published by the Free Software
    Foundation; either version 2 of the License, or (at your option) any
    later version.

or both in parallel, as here.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received copies of the GNU General Public License and the
GNU Lesser General Public License along with the GNU MP Library.  If not,
see https://www.gnu.org/licenses/.  */

#define GMP_LIMB_BITS 64
#define GMP_LIMB_BYTES 8

/* Disable use of slow functions.  FIXME: We should disable lib inclusion.  */
#undef HAVE_NATIVE_mpn_mul_2
#undef HAVE_NATIVE_mpn_addmul_2

/* 2400 MHz Intel Atom C2758 Silvermont/Rangeley */
/* FFT tuning limit = 468153400 */
/* Generated by tuneup.c, 2019-10-19, gcc 8.3 */

#define MOD_1_NORM_THRESHOLD                 0  /* always */
#define MOD_1_UNNORM_THRESHOLD               0  /* always */
#define MOD_1N_TO_MOD_1_1_THRESHOLD          4
#define MOD_1U_TO_MOD_1_1_THRESHOLD          3
#define MOD_1_1_TO_MOD_1_2_THRESHOLD        55
#define MOD_1_2_TO_MOD_1_4_THRESHOLD         0  /* never mpn_mod_1s_2p */
#define PREINV_MOD_1_TO_MOD_1_THRESHOLD      9
#define USE_PREINV_DIVREM_1                  1  /* native */
#define DIV_QR_1_NORM_THRESHOLD              1
#define DIV_QR_1_UNNORM_THRESHOLD        MP_SIZE_T_MAX  /* never */
#define DIV_QR_2_PI2_THRESHOLD           MP_SIZE_T_MAX  /* never */
#define DIVEXACT_1_THRESHOLD                 0  /* always (native) */
#define BMOD_1_TO_MOD_1_THRESHOLD        MP_SIZE_T_MAX  /* never */

#define DIV_1_VS_MUL_1_PERCENT             168

#define MUL_TOOM22_THRESHOLD                19
#define MUL_TOOM33_THRESHOLD                66
#define MUL_TOOM44_THRESHOLD               152
#define MUL_TOOM6H_THRESHOLD               222
#define MUL_TOOM8H_THRESHOLD               333

#define MUL_TOOM32_TO_TOOM43_THRESHOLD     105
#define MUL_TOOM32_TO_TOOM53_THRESHOLD     114
#define MUL_TOOM42_TO_TOOM53_THRESHOLD     105
#define MUL_TOOM42_TO_TOOM63_THRESHOLD     113
#define MUL_TOOM43_TO_TOOM54_THRESHOLD      88

#define SQR_BASECASE_THRESHOLD               0  /* always (native) */
#define SQR_TOOM2_THRESHOLD                 24
#define SQR_TOOM3_THRESHOLD                 97
#define SQR_TOOM4_THRESHOLD                232
#define SQR_TOOM6_THRESHOLD                286
#define SQR_TOOM8_THRESHOLD                  0  /* always */

#define MULMID_TOOM42_THRESHOLD             24

#define MULMOD_BNM1_THRESHOLD               13
#define SQRMOD_BNM1_THRESHOLD               15

#define MUL_FFT_MODF_THRESHOLD             340  /* k = 5 */
#define MUL_FFT_TABLE3                                      \
  { {    340, 5}, {     17, 6}, {      9, 5}, {     19, 6}, \
    {     17, 7}, {      9, 6}, {     20, 7}, {     11, 6}, \
    {     23, 7}, {     17, 8}, {      9, 7}, {     21, 8}, \
    {     11, 7}, {     23, 8}, {     13, 7}, {     27, 8}, \
    {     15, 7}, {     31, 8}, {     21, 9}, {     11, 8}, \
    {     27, 9}, {     15, 8}, {     33, 9}, {     19, 8}, \
    {     39, 9}, {     23, 8}, {     47, 9}, {     27,10}, \
    {     15, 9}, {     39,10}, {     23, 9}, {     51,11}, \
    {     15,10}, {     31, 9}, {     67,10}, {     39, 9}, \
    {     79,10}, {     47, 9}, {     95,10}, {     55,11}, \
    {     31,10}, {     79,11}, {     47,10}, {     95,12}, \
    {     31,11}, {     63,10}, {    127, 9}, {    255,10}, \
    {    135,11}, {     79, 9}, {    319,11}, {     95,10}, \
    {    191, 9}, {    383,10}, {    207, 9}, {    415,11}, \
    {    111,12}, {     63,11}, {    127,10}, {    255, 9}, \
    {    511,10}, {    271,11}, {    143,10}, {    287, 9}, \
    {    575,10}, {    303,11}, {    159,10}, {    319,12}, \
    {     95,11}, {    191,10}, {    383,11}, {    207,10}, \
    {    415,13}, {     63,12}, {    127,11}, {    255,10}, \
    {    511,11}, {    271,10}, {    543,11}, {    287,10}, \
    {    575,11}, {    319,10}, {    639,11}, {    351,10}, \
    {    703, 9}, {   1407,12}, {    191,11}, {    415,10}, \
    {    831,12}, {    223,11}, {    479,13}, {    127,12}, \
    {    255,11}, {    543,12}, {    287,11}, {    575,10}, \
    {   1151,12}, {    319,11}, {    639,12}, {    351,11}, \
    {    703,10}, {   1407,13}, {    191,12}, {    415,11}, \
    {    831,10}, {   1663,12}, {    479,14}, {    127,13}, \
    {    255,12}, {    543,11}, {   1087,10}, {   2175,12}, \
    {    575,11}, {   1151,13}, {    319,12}, {    639,11}, \
    {   1279,12}, {    703,11}, {   1407,13}, {    383,12}, \
    {    831,11}, {   1663,13}, {    447,12}, {    959,14}, \
    {    255,13}, {    511,12}, {   1087,11}, {   2175,13}, \
    {    575,12}, {   1215,11}, {   2431,10}, {   4863,13}, \
    {    639,12}, {   1279,13}, {    703,12}, {   1407,14}, \
    {    383,13}, {    831,12}, {   1663,13}, {    959,15}, \
    {    255,14}, {    511,13}, {   1087,12}, {   2175,13}, \
    {   1215,12}, {   2431,11}, {   4863,14}, {    639,13}, \
    {   1407,12}, {   2815,13}, {   1471,12}, {   2943,11}, \
    {   5887,14}, {    767,13}, {   1663,14}, {    895,13}, \
    {   1919,15}, {    511,14}, {   1023,13}, {   2175,14}, \
    {   1151,13}, {   2431,12}, {   4863,14}, {   1279,13}, \
    {   2559,14}, {   1407,13}, {   2943,12}, {   5887,15}, \
    {    767,14}, {   1663,13}, {   3455,12}, {   6911,14}, \
    {   1919,16}, {    511,15}, {   1023,14}, {   2431,13}, \
    {   4863,15}, {   1279,14}, {   2943,13}, {   5887,12}, \
    {  11775,15}, {   1535,14}, {   3455,13}, {   6911,15}, \
    {   1791,14}, {   3839,13}, {   7679,16}, {   1023,15}, \
    {   2047,14}, {   4223,15}, {   2303,14}, {   4863,15}, \
    {   2815,14}, {   5887,13}, {  11775,16}, {   1535,15}, \
    {   3327,14}, {   6911,15}, {   3839,14}, {   7679,17}, \
    {   1023,16}, {   2047,15}, {   4863,16}, {   2559,15}, \
    {   5887,14}, {  11775,16}, {   3071,15}, {   6911,16}, \
    {   3583,15}, {   7679,14}, {  15359,17}, {   2047,16}, \
    {   4607,15}, {   9215,16}, {   5631,15}, {  11775,17}, \
    {   3071,16}, {  65536,17}, { 131072,18}, { 262144,19}, \
    { 524288,20}, {1048576,21}, {2097152,22}, {4194304,23}, \
    {8388608,24} }
#define MUL_FFT_TABLE3_SIZE 225
#define MUL_FFT_THRESHOLD                 3712

#define SQR_FFT_MODF_THRESHOLD             308  /* k = 5 */
#define SQR_FFT_TABLE3                                      \
  { {    308, 5}, {     17, 6}, {      9, 5}, {     19, 6}, \
    {     21, 7}, {     11, 6}, {     23, 7}, {     17, 8}, \
    {      9, 7}, {     21, 8}, {     11, 7}, {     23, 8}, \
    {     13, 7}, {     27, 8}, {     15, 7}, {     31, 8}, \
    {     21, 9}, {     11, 8}, {     27, 9}, {     15, 8}, \
    {     33, 9}, {     19, 8}, {     39, 9}, {     23, 8}, \
    {     47, 9}, {     27,10}, {     15, 9}, {     39,10}, \
    {     23, 9}, {     47,11}, {     15,10}, {     31, 9}, \
    {     63,10}, {     39, 9}, {     79,10}, {     47, 9}, \
    {     95,11}, {     31,10}, {     79,11}, {     47,10}, \
    {     95,12}, {     31,11}, {     63,10}, {    127, 9}, \
    {    255, 8}, {    511, 9}, {    271, 8}, {    543,11}, \
    {     79,10}, {    159, 9}, {    319, 8}, {    639,10}, \
    {    175,11}, {     95,10}, {    191, 9}, {    383,10}, \
    {    207, 9}, {    415,12}, {     63,11}, {    127,10}, \
    {    255, 9}, {    511,10}, {    271, 9}, {    543,10}, \
    {    287, 9}, {    575,10}, {    303,11}, {    159,10}, \
    {    319, 9}, {    639,11}, {    175,10}, {    351,12}, \
    {     95,11}, {    191,10}, {    383,11}, {    207,10}, \
    {    415,13}, {     63,12}, {    127,11}, {    255,10}, \
    {    511,11}, {    271,10}, {    543,11}, {    287,10}, \
    {    575,11}, {    303,12}, {    159,11}, {    319,10}, \
    {    639,11}, {    351,10}, {    703,12}, {    191,11}, \
    {    383,10}, {    767,11}, {    415,10}, {    831,12}, \
    {    223,11}, {    479,12}, {    255,11}, {    543,12}, \
    {    287,11}, {    575,12}, {    319,11}, {    639,12}, \
    {    351,11}, {    703,10}, {   1407,13}, {    191,12}, \
    {    383,11}, {    767,12}, {    415,11}, {    831,12}, \
    {    479,13}, {    255,12}, {    543,11}, {   1087,10}, \
    {   2175,12}, {    575,11}, {   1151,12}, {    607,13}, \
    {    319,12}, {    639,11}, {   1279,12}, {    703,11}, \
    {   1407,13}, {    383,12}, {    831,11}, {   1663,13}, \
    {    447,12}, {    895,14}, {    255,13}, {    511,12}, \
    {   1087,11}, {   2175,13}, {    575,12}, {   1215,11}, \
    {   2431,13}, {    639,12}, {   1279,13}, {    703,12}, \
    {   1407,14}, {    383,13}, {    831,12}, {   1663,13}, \
    {    959,15}, {    255,14}, {    511,13}, {   1087,12}, \
    {   2175,13}, {   1215,12}, {   2431,14}, {    639,13}, \
    {   1343,12}, {   2687,13}, {   1407,12}, {   2815,13}, \
    {   1471,12}, {   2943,14}, {    767,13}, {   1663,14}, \
    {    895,13}, {   1791,15}, {    511,14}, {   1023,13}, \
    {   2175,14}, {   1151,13}, {   2431,12}, {   4863,14}, \
    {   1279,13}, {   2687,14}, {   1407,13}, {   2943,15}, \
    {    767,14}, {   1663,13}, {   3455,12}, {   6911,14}, \
    {   1791,13}, {   3583,16}, {    511,15}, {   1023,14}, \
    {   2431,13}, {   4863,15}, {   1279,14}, {   2943,13}, \
    {   5887,12}, {  11775,15}, {   1535,14}, {   3455,13}, \
    {   6911,15}, {   1791,14}, {   3839,13}, {   7679,16}, \
    {   1023,15}, {   2047,14}, {   4223,15}, {   2303,14}, \
    {   4863,15}, {   2815,14}, {   5887,13}, {  11775,16}, \
    {   1535,15}, {   3071,14}, {   6143,15}, {   3327,14}, \
    {   6911,15}, {   3839,14}, {   7679,17}, {   1023,16}, \
    {   2047,15}, {   4863,16}, {   2559,15}, {   5887,14}, \
    {  11775,16}, {   3071,15}, {   6911,16}, {   3583,15}, \
    {   7679,14}, {  15359,17}, {   2047,16}, {   4607,15}, \
    {   9983,16}, {   5631,15}, {  11775,17}, {   3071,16}, \
    {  65536,17}, { 131072,18}, { 262144,19}, { 524288,20}, \
    {1048576,21}, {2097152,22}, {4194304,23}, {8388608,24} }
#define SQR_FFT_TABLE3_SIZE 232
#define SQR_FFT_THRESHOLD                 2752

#define MULLO_BASECASE_THRESHOLD             0  /* always */
#define MULLO_DC_THRESHOLD                  55
#define MULLO_MUL_N_THRESHOLD             6633
#define SQRLO_BASECASE_THRESHOLD             9
#define SQRLO_DC_THRESHOLD                   0  /* never mpn_sqrlo_basecase */
#define SQRLO_SQR_THRESHOLD               5397

#define DC_DIV_QR_THRESHOLD                 33
#define DC_DIVAPPR_Q_THRESHOLD             222
#define DC_BDIV_QR_THRESHOLD                31
#define DC_BDIV_Q_THRESHOLD                147

#define INV_MULMOD_BNM1_THRESHOLD           37
#define INV_NEWTON_THRESHOLD               222
#define INV_APPR_THRESHOLD                 222

#define BINV_NEWTON_THRESHOLD              212
#define REDC_1_TO_REDC_2_THRESHOLD          55
#define REDC_2_TO_REDC_N_THRESHOLD           0  /* always */

#define MU_DIV_QR_THRESHOLD               1142
#define MU_DIVAPPR_Q_THRESHOLD            1142
#define MUPI_DIV_QR_THRESHOLD               81
#define MU_BDIV_QR_THRESHOLD               942
#define MU_BDIV_Q_THRESHOLD               1043

#define POWM_SEC_TABLE  1,34,102,588,1730

#define GET_STR_DC_THRESHOLD                17
#define GET_STR_PRECOMPUTE_THRESHOLD        30
#define SET_STR_DC_THRESHOLD               381
#define SET_STR_PRECOMPUTE_THRESHOLD      1659

#define FAC_DSC_THRESHOLD                  351
#define FAC_ODD_THRESHOLD                   27

#define MATRIX22_STRASSEN_THRESHOLD         16
#define HGCD2_DIV1_METHOD                    3  /* 3.06% faster than 1 */
#define HGCD_THRESHOLD                     120
#define HGCD_APPR_THRESHOLD                153
#define HGCD_REDUCE_THRESHOLD             2121
#define GCD_DC_THRESHOLD                   416
#define GCDEXT_DC_THRESHOLD                309
#define JACOBI_BASE_METHOD                   1  /* 2.28% faster than 3 */

/* Tuneup completed successfully, took 938046 seconds */
